@@ -14,7 +14,7 @@ import pyarrow.csv
 from packaging.version import parse
 
 from chalk.features import Feature
-from chalk.integrations.named import load_integration_variable
+from chalk.integrations.named import create_integration_variable, load_integration_variable
 from chalk.sql._internal.query_execution_parameters import QueryExecutionParameters
 from chalk.sql._internal.sql_source import (
     BaseSQLSource,
@@ -28,6 +28,7 @@ from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
 from chalk.utils.environment_parsing import env_var_bool
 from chalk.utils.log_with_context import get_logger
 from chalk.utils.missing_dependency import missing_dependency_exception
+from chalk.utils.tracing import safe_add_metrics, safe_add_tags, safe_trace
 
 if TYPE_CHECKING:
     import polars as pl
@@ -67,37 +68,12 @@ else:
 _logger = get_logger(__name__)
 
 
-@contextlib.contextmanager
-def _safe_trace(span_id: str):
-    try:
-        import ddtrace
-
-        with ddtrace.tracer.trace(name=span_id):
-            yield
-    except ImportError:
-        yield
-
-
-def _safe_add_metrics(metrics: Mapping[str, Union[int, float]]):
-    try:
-        import ddtrace
-
-        span = ddtrace.tracer.current_span()
-        if span:
-            span.set_metrics(dict(metrics))
-    except ImportError:
-        ...
-
-
-def _safe_add_tags(tags: Mapping[str, str]):
-    try:
-        import ddtrace
-
-        span = ddtrace.tracer.current_span()
-        if span:
-            span.set_tags(dict(tags))
-    except ImportError:
-        ...
+_PGHOST_NAME = "PGHOST"
+_PGPORT_NAME = "PGPORT"
+_PGDATABASE_NAME = "PGDATABASE"
+_PGUSER_NAME = "PGUSER"
+_PGPASSWORD_NAME = "PGPASSWORD"
+_ENGINE_ARGUMENTS_NAME = "ENGINE_ARGUMENTS"
 
 
 class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
@@ -113,6 +89,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
         name: Optional[str] = None,
         engine_args: Optional[Dict[str, Any]] = None,
         async_engine_args: Optional[Dict[str, Any]] = None,
+        integration_variable_override: Optional[Mapping[str, str]] = None,
     ):
         try:
             import psycopg  # Used for the async driver
@@ -131,15 +108,25 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
                 "postgresql.psycopg_async", "chalk.sql._internal.integrations.psycopg3.psycopg_dialect", "dialect_async"
             )
         self.name = name
-        self.host = host or load_integration_variable(integration_name=name, name="PGHOST")
+        self.host = host or load_integration_variable(
+            integration_name=name, name=_PGHOST_NAME, override=integration_variable_override
+        )
         self.port = (
             int(port)
             if port is not None
-            else load_integration_variable(integration_name=name, name="PGPORT", parser=int)
+            else load_integration_variable(
+                integration_name=name, name=_PGPORT_NAME, parser=int, override=integration_variable_override
+            )
+        )
+        self.db = db or load_integration_variable(
+            integration_name=name, name=_PGDATABASE_NAME, override=integration_variable_override
+        )
+        self.user = user or load_integration_variable(
+            integration_name=name, name=_PGUSER_NAME, override=integration_variable_override
+        )
+        self.password = password or load_integration_variable(
+            integration_name=name, name=_PGPASSWORD_NAME, override=integration_variable_override
         )
-        self.db = db or load_integration_variable(integration_name=name, name="PGDATABASE")
-        self.user = user or load_integration_variable(integration_name=name, name="PGUSER")
-        self.password = password or load_integration_variable(integration_name=name, name="PGPASSWORD")
         self.ingested_tables: Dict[str, Any] = {}
         if engine_args is None:
             engine_args = {}
@@ -147,7 +134,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
             async_engine_args = {}
 
         if name:
-            engine_args_from_ui = self._load_env_engine_args(name)
+            engine_args_from_ui = self._load_env_engine_args(name, override=integration_variable_override)
             for k, v in engine_args_from_ui.items():
                 engine_args.setdefault(k, v)
                 async_engine_args.setdefault(k, v)
@@ -210,7 +197,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
         columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         query_execution_parameters: QueryExecutionParameters,
     ) -> pyarrow.Table:
-        with _safe_trace("_get_parsed_table_via_polars"):
+        with safe_trace("_get_parsed_table_via_polars"):
             import polars as pl
 
             buffer = self._get_result_csv_buffer(
@@ -236,31 +223,31 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
             datetime_col_to_dtype: Dict[str, pl.Datetime] = {}
             for col_name, feature in features.items():
                 dtype = feature.converter.polars_dtype
-                if dtype == pl.Boolean:
+                if dtype == pl.Boolean:  # pyright: ignore[reportUnnecessaryComparison]
                     # Parsing "t" and "f" as booleans fails.
                     # We will parse the column as a string, and then
                     # convert it to a boolean later.
                     parse_dtypes[col_name] = pl.Utf8
                     boolean_columns.append(col_name)
-                elif dtype == pl.List:
+                elif dtype == pl.List:  # pyright: ignore[reportUnnecessaryComparison]
                     # We will parse the list as a string, and then
                     # convert it to a list later. Polars does not
                     # support parsing lists directly from CSV.
                     parse_dtypes[col_name] = pl.Utf8
-                elif dtype == pl.Datetime:
+                elif dtype == pl.Datetime:  # pyright: ignore[reportUnnecessaryComparison]
                     # Parse as str, convert to datetime later.
                     # It's important to parse datetimes as strings,
                     # because `polars.read_csv()` just turns
-                    # unparseable datetimes into nulls, a huge no-no.
+                    # unparsable datetimes into nulls, a huge no-no.
                     parse_dtypes[col_name] = pl.Utf8
                     # Make linter happy
                     dtype = cast(pl.Datetime, dtype)
                     datetime_col_to_dtype[col_name] = dtype
-                elif dtype == pl.Date:
+                elif dtype == pl.Date:  # pyright: ignore[reportUnnecessaryComparison]
                     # Parse as str, convert to date later.
                     # It's important to parse dates as strings,
                     # because `polars.read_csv()` just turns
-                    # unparseable dates into nulls, a huge no-no.
+                    # unparsable dates into nulls, a huge no-no.
                     parse_dtypes[col_name] = pl.Utf8
                     date_columns.append(col_name)
                 else:
@@ -274,7 +261,9 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
             #
             # pl.read_csv(use_pyarrow=True) has the same performance degradation,
             # UNLESS a `dtypes` arg is provided.
-            pl_table = pl.read_csv(buffer, dtypes=parse_dtypes)
+
+            # 'dtypes' deprecated for 'schema_overrides' in polars 0.20+, but parameter renamed without breaking
+            pl_table = pl.read_csv(buffer, dtypes=parse_dtypes)  # pyright: ignore[reportCallIssue]
             if boolean_columns:
                 # DO NOT use map_dict. Causes a segfault when multiple uvicorn workers are handling
                 # requests in parallel.
@@ -321,7 +310,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
 
             parsed_table = pl_table.to_arrow()
 
-            _safe_add_metrics(
+            safe_add_metrics(
                 {
                     "postgresl_polars_row_count": len(parsed_table),
                     "postgresl_polars_bytes": parsed_table.nbytes,
@@ -336,7 +325,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
         escape_char: str,
         query_execution_parameters: QueryExecutionParameters,
     ) -> io.BytesIO:
-        with _safe_trace("_get_result_csv_buffer"):
+        with safe_trace("_get_result_csv_buffer"):
             import psycopg2.sql
 
             stmt, positional_params, named_params = self.compile_query(finalized_query, paramstyle="named")
@@ -353,7 +342,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
                 env_var_bool("CHALK_SKIP_PG_DATETIME_ZONE_CAST")
                 or query_execution_parameters.postgres.skip_datetime_timezone_cast
             ):
-                with _safe_trace("skip_datetime_timezone_cast"):
+                with safe_trace("skip_datetime_timezone_cast"):
                     formatted_stmt = psycopg2.sql.SQL(stmt).format(
                         **{
                             k: psycopg2.sql.Literal(str(v)) if isinstance(v, datetime) else psycopg2.sql.Literal(v)
@@ -368,11 +357,11 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
                 dbapi = cnx.connection
                 with dbapi.cursor() as cursor:
                     buffer = io.BytesIO()
-                    with _safe_trace("copy_expert"):
-                        _safe_add_tags({"query": formatted_stmt.as_string(cursor)[0:2000]})
+                    with safe_trace("copy_expert"):
+                        safe_add_tags({"query": formatted_stmt.as_string(cursor)[0:2000]})
                         cursor.copy_expert(formatted_stmt, buffer)
                         written_bytes = buffer.getbuffer().nbytes
-                        _safe_add_metrics({"postgresl_bytes": written_bytes})
+                        safe_add_metrics({"postgresl_bytes": written_bytes})
 
             buffer.seek(0)
 
@@ -385,7 +374,7 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
         connection: Optional[Connection],
         query_execution_parameters: QueryExecutionParameters,
     ) -> Iterable[pa.RecordBatch]:
-        with _safe_trace("postgres.execute_query_efficient"):
+        with safe_trace("postgres.execute_query_efficient"):
             from sqlalchemy.sql import Select
 
             if finalized_query.finalizer in (Finalizer.FIRST, Finalizer.ONE, Finalizer.ONE_OR_NONE):
@@ -444,6 +433,19 @@ class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIn
                 yield pa.RecordBatch.from_pydict({k: [] for k in parsed_table.schema.names}, parsed_table.schema)
             return
 
+    def _recreate_integration_variables(self) -> dict[str, str]:
+        return {
+            k: v
+            for k, v in [
+                create_integration_variable(_PGHOST_NAME, self.name, self.host),
+                create_integration_variable(_PGPORT_NAME, self.name, self.port),
+                create_integration_variable(_PGDATABASE_NAME, self.name, self.db),
+                create_integration_variable(_PGUSER_NAME, self.name, self.user),
+                create_integration_variable(_PGPASSWORD_NAME, self.name, self.password),
+            ]
+            if v is not None
+        }
+
 
 def _reformat_compiled_query_string_for_csv_output(
     stmt: str,
@@ -516,6 +518,7 @@ def _get_df_with_parsed_dt_column(
         target_dtype = cast(pl.Datetime, target_dtype)
         original_target_tz = target_dtype.time_zone
         if use_new_parsing:
+            assert target_dtype.time_unit is not None, f"polars dtype {target_dtype} incorrectly set with no time unit."
             parsing_dtype = pl.Datetime(time_zone="UTC", time_unit=target_dtype.time_unit)
         else:
             target_dtype = pl.Datetime
