@@ -1,7 +1,8 @@
-import dataclasses
+from __future__ import annotations
+
 import os
-from dataclasses import dataclass
-from typing import Optional, Set
+from dataclasses import dataclass, field
+from typing import Optional
 
 from chalk.utils.environment_parsing import env_var_bool
 
@@ -37,6 +38,14 @@ class SnowflakeQueryExecutionParameters:
     """
 
 
+@dataclass(frozen=True)
+class BigqueryQueryExecutionParameters:
+    bigquery_unload_path: Optional[str]
+    """
+    The name of the GCS or S3 bucket to use for unloading data. If None, queries will be executed to a temp table.
+    """
+
+
 @dataclass(frozen=True)
 class QueryExecutionParameters:
     attempt_efficient_execution: bool
@@ -49,6 +58,8 @@ class QueryExecutionParameters:
 
     snowflake: SnowflakeQueryExecutionParameters
 
+    bigquery: BigqueryQueryExecutionParameters
+
     yield_empty_batches: bool = False
     """Whether to yield empty batches. This can be useful to capture the schema of an otherwise-empty results set"""
 
@@ -62,7 +73,7 @@ class QueryExecutionParameters:
     num_client_prefetch_threads: int = 4
     """Number of threads to use when downloading query results."""
 
-    tags: Set[str] = dataclasses.field(default_factory=set)
+    tags: frozenset[str] = field(default_factory=frozenset)
     """Tags that can be used to select a source. These tags are proxied from the overlying Chalk query context."""
 
 
@@ -73,6 +84,8 @@ def query_execution_parameters_from_env_vars():
     """
     return QueryExecutionParameters(
         attempt_efficient_execution=not env_var_bool("CHALK_FORCE_SQLALCHEMY_QUERY_EXECUTION"),
+        max_prefetch_size_bytes=int(os.getenv("CHALK_MAX_PREFETCH_SIZE_BYTES", "1073741824")),
+        num_client_prefetch_threads=int(os.getenv("CHALK_NUM_CLIENT_PREFETCH_THREADS", "4")),
         postgres=PostgresQueryExecutionParameters(
             attempt_efficient_postgres_execution=True,
             polars_read_csv=env_var_bool("CHALK_FORCE_POLARS_READ_CSV"),
@@ -81,4 +94,7 @@ def query_execution_parameters_from_env_vars():
         snowflake=SnowflakeQueryExecutionParameters(
             snowflake_unload_stage=os.getenv("CHALK_SNOWFLAKE_UNLOAD_STAGE", None),
         ),
+        bigquery=BigqueryQueryExecutionParameters(
+            bigquery_unload_path=os.getenv("CHALK_BIGQUERY_UNLOAD_PATH", None),
+        ),
     )
