@@ -2,6 +2,7 @@ from __future__ import annotations
 
 import dataclasses
 import warnings
+from collections import defaultdict
 from enum import Enum
 from typing import (
     TYPE_CHECKING,
@@ -49,25 +50,43 @@ class Finalizer(str, Enum):
     ALL = "All"
 
 
-def _get_matching_root_fqn(normalized_col_name: str, expected_feature_root_fqns: Collection[str]) -> Optional[str]:
+def _get_matching_root_fqn(
+    normalized_to_original_col: Mapping[str, str], expected_feature_root_fqns: Collection[str]
+) -> dict[str, str]:
+    # Returns a map from original_col to the matched columns
     candidates: List[str] = []
+    lookup: defaultdict[str, list[str]] = defaultdict(list)
     for x in expected_feature_root_fqns:
         root_fqn_normalized = normalize_string_for_matching(x.lower())
         without_root_ns = normalize_string_for_matching(".".join(x.split(".")[1:]))
-        if normalized_col_name == root_fqn_normalized or normalized_col_name == without_root_ns:
-            candidates.append(x)
-    if len(candidates) == 0:
-        return None
-    if len(candidates) > 1:
-        # We really shouldn't hit this, unless if features are case-sensitive and the user is querying
-        # snowflake which is case-insensitive
-        raise ValueError(
-            (
-                f"Column '{normalized_col_name}' was ambiguous which feature it referred to. "
-                f"Possible candidates were: {candidates}"
+        lookup[root_fqn_normalized].append(x)
+        lookup[without_root_ns].append(x)
+        if len(x.split("@")) == 2:
+            non_versioned = x.split("@")[0]
+            non_versioned_normalized = normalize_string_for_matching(non_versioned.lower())
+            non_versioned_without_root_ns = normalize_string_for_matching(".".join(non_versioned.split(".")[1:]))
+            lookup[non_versioned_normalized].append(x)
+            lookup[non_versioned_without_root_ns].append(x)
+
+    new_mapped_cols = {}
+    for normalized_col_name, original_col_name in normalized_to_original_col.items():
+        if len(original_col_name.split("@")) == 2:
+            version_number = original_col_name.split("@")[1]
+            normalized_col_name = normalized_col_name[: -len(f"{version_number}")]
+        candidates = lookup[normalized_col_name]
+        if len(candidates) == 0:
+            continue
+        if len(candidates) > 1:
+            # We really shouldn't hit this, unless if features are case-sensitive and the user is querying
+            # snowflake which is case-insensitive
+            raise ValueError(
+                (
+                    f"Column '{normalized_col_name}' was ambiguous which feature it referred to. "
+                    f"Possible candidates were: {candidates}"
+                )
             )
-        )
-    return candidates[0]
+        new_mapped_cols[original_col_name] = candidates[0]
+    return new_mapped_cols
 
 
 class FinalizedChalkQuery:
@@ -100,6 +119,13 @@ class FinalizedChalkQuery:
         self._temp_tables = dict(temp_tables) if temp_tables is not None else {}
         self._is_empty = is_empty
 
+    @classmethod
+    def __init_subclass__(cls, **kwargs: Any):
+        super().__init_subclass__(**kwargs)
+        raise RuntimeError(
+            "FinalizedChalkQuery should not be subclassed. The Chalk runtime assumes that this class is final."
+        )
+
     @property
     def incremental_settings(self):
         return self._incremental_settings
@@ -254,7 +280,7 @@ class FinalizedChalkQuery:
     def _check_incremental_settings(self):
         if self.incremental_settings is not None:
             # FIXME: Move the incrementalization logic here, so then the `execute` and `execute_to_dataframe`
-            # methods can take the hwm timestamp as a paramater, to allow for direct execution
+            # methods can take the hwm timestamp as a parameter, to allow for direct execution
             warnings.warn(
                 (
                     "This query specified an incremental configuration, which has not been applied. "
@@ -560,11 +586,12 @@ class FinalizedChalkQuery:
             unexpected_fields = [x for x in self._fields.values() if x.root_fqn not in expected_features]
             for x in unexpected_fields:
                 del ans[normalized_to_original_col[normalize_string_for_matching(x.root_fqn)]]
-            for normalized_col_name, original_col_name in normalized_to_original_col.items():
-                if original_col_name not in ans:
-                    maybe_matching_root_fqn = _get_matching_root_fqn(normalized_col_name, expected_features)
-                    if maybe_matching_root_fqn is not None:
-                        ans[original_col_name] = maybe_matching_root_fqn
+            unsolved_normalized_to_original_col = {
+                normalized_col: original_col
+                for normalized_col, original_col in normalized_to_original_col.items()
+                if original_col not in ans
+            }
+            ans.update(_get_matching_root_fqn(unsolved_normalized_to_original_col, expected_features))
         return ans
 
 
